/***************************************************************************
 *																		 *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or	 *
 *   (at your option) any later version.								   *
 *																		 *
 ***************************************************************************/

#include "status.h"
#include "misc.h"
#include "protocol.h"
#include "kadu.h"
#include "userlist.h"
#include "config_file.h"
#include "icons_manager.h"

#include "desc_history.h"
#include "descriptions_window.h"

#define MODULE_DESC_HIST_VERSION 1.1

DescHistory* descHistory;

extern "C" int desc_history_init()
{
	descHistory = new DescHistory();
	int res = descHistory->initialize();
	if (res)
		delete descHistory;

	MainConfigurationWindow::registerUiFile(dataPath("kadu/modules/configuration/desc_history.ui"), descHistory);
	return res;
}

extern "C" void desc_history_close()
{
	MainConfigurationWindow::unregisterUiFile(dataPath("kadu/modules/configuration/desc_history.ui"), descHistory);
	delete descHistory;
}

DescHistory::DescHistory()
{
	connect(userlist, SIGNAL(statusChanged(UserListElement, QString, const UserStatus &, bool, bool)),
		this, SLOT(statusChanged(UserListElement, QString, const UserStatus &, bool, bool)));

	createDefaultConfiguration();
	menuItemId = kadu->mainMenu()->insertItem(icons_manager->loadIcon("DescHistory"), tr("Descriptions history"), this, SLOT(showDescHistoryWin()), CTRL+Key_D, -1, kadu->mainMenu()->count() - 6);
	userMenuItemId = UserBox::userboxmenu->addItemAtPos( 11, tr("Descriptions history"), this, SLOT(showDescHistoryWinSelected()) );

	// Test SELECT
/*
	char *zErrMsg = 0;
	res = sqlite3_exec( db, "SELECT * from t", callback, 0, &zErrMsg );
	if( res != SQLITE_OK )
	{
		errMsg = tr("SQL error: %1").arg(zErrMsg);
		sqlite3_free( zErrMsg );
		return;
	}
*/
}

DescHistory::~DescHistory()
{
	if ( db != 0 )
	{
		if ( sqlite3_close(db) == SQLITE_BUSY )
		{
			sqlite3_interrupt(db);
			sqlite3_close(db);
		}
	}

	kadu->mainMenu()->removeItem(menuItemId);
	UserBox::userboxmenu->removeItem(userMenuItemId);

	disconnect(userlist, SIGNAL(statusChanged(UserListElement, QString, const UserStatus &, bool, bool)),
		this, SLOT(statusChanged(UserListElement, QString, const UserStatus &, bool, bool)));
}

int DescHistory::initialize()
{
	int res = sqlite3_open( ggPath("desc_history.db"), &db );
	if ( res )
	{
		//errMsg = tr("Can't open %1 database: %2").arg(ggPath("desc_history.db")).arg(sqlite3_errmsg( db ));
		sqlite3_close( db );
		return 1;
	}

	QStringList params;
	QString result = oneColumn("SELECT count(*) AS cnt FROM sqlite_master WHERE type = 'table' AND name = 'history';", params);
	if ( result != "1" )
	{
		call("CREATE TABLE [history] ([id] INTEGER PRIMARY KEY AUTOINCREMENT, [created_on] DATE NOT NULL, [uid] TEXT NOT NULL, [description] TEXT, [marked] BOOLEAN NOT NULL DEFAULT 0);", params);
		call("CREATE INDEX [idx_history] ON [history] ([uid]);", params);
		call("CREATE INDEX [idx_history_date] ON [history] ([created_on]);", params);
		call("CREATE INDEX [idx_history_date_uid] ON [history] ([uid], [created_on]);", params);
	}

	bool exclude = config_file.readBoolEntry( "DescHist", "ExcludeMarked", true );
	int limit = config_file.readNumEntry( "DescHist", "MaxDescriptions", 300 );

	QString sql = "SELECT count(*) AS cnt, uid FROM history%1 GROUP BY uid ORDER BY MAX (id);";
	if (exclude)
		sql = sql.arg(" WHERE marked <> 1");
	else
		sql = sql.arg("");

	QValueList< QMap<QString,QString> > results = query(sql, params);
	for ( QValueList< QMap<QString,QString> >::Iterator it = results.begin(); it != results.end(); it++ )
	{
		QMap<QString,QString> row = *it;
		int cnt = row["cnt"].toInt();
		if ( cnt > limit )
		{
			params.clear();
			params += row["uid"];
			params += QString::number(limit);
			QString id = oneColumn("SELECT id FROM history WHERE uid = ? ORDER BY id DESC LIMIT 1 OFFSET ?;", params);
			params.clear();
			params += row["uid"];
			params += id;
			call("DELETE FROM history WHERE uid = ? AND id <= ?;", params);
		}
	}


	return 0;
}

void DescHistory::call(QString sql, QStringList params)
{
	sqlite3_stmt* stmt;
	const char *pzTail;
	int rescode = sqlite3_prepare_v2(db, sql.utf8(), sql.length(), &stmt, &pzTail);
	if ( rescode != SQLITE_OK )
	{
		printf("[DescHistory module] Error while calling '%s':\n%s\n", sql.latin1(), sqlite3_errmsg(db));
		return;
	}

	for ( uint i = 0; i < params.count(); i++ )
	{
		rescode = sqlite3_bind_text(stmt, i+1, params[i].utf8(), params[i].utf8().length(), SQLITE_TRANSIENT);
		if (rescode != SQLITE_OK)
		{
			printf("[DescHistory module] Error binding parameter %d for sql: %s\nError message: %s\n", i, sql.latin1(), sqlite3_errmsg(db));
		}
	}
	//.replace("'", "''")

	rescode = sqlite3_step(stmt);
	if (rescode == SQLITE_ERROR || rescode == SQLITE_MISUSE)
		printf("[DescHistory module] Error while calling '%s':\n%s\n", sql.latin1(), sqlite3_errmsg(db));
// 	else
// 		printf("Execution rescode is %d (done=%d, row=%d, err=%d)\n", rescode, SQLITE_DONE, SQLITE_ROW, SQLITE_ERROR);

	sqlite3_finalize(stmt);
}

QValueList< QMap<QString,QString> > DescHistory::query(QString sql, QStringList params)
{
	sqlite3_stmt* stmt;
	const char *pzTail;
	int rescode;
	QValueList< QMap<QString,QString> > results;

	sqlite3_prepare_v2(db, sql.utf8(), sql.length(), &stmt, &pzTail);

	for ( uint i = 0; i < params.count(); i++ )
	{
		rescode = sqlite3_bind_text(stmt, i+1, params[i].utf8(), params[i].utf8().length(), SQLITE_TRANSIENT);
		if (rescode != SQLITE_OK)
		{
			printf("[DescHistory module] Error binding parameter %d for sql: %s\nError message: %s\n", i, sql.latin1(), sqlite3_errmsg(db));
		}
	}

	int res = sqlite3_step(stmt);
	int columns = sqlite3_column_count(stmt);

	while ( res == SQLITE_ROW )
	{
		QMap<QString,QString> row;
		for ( int c = 0; c < columns; c++ )
		{
			QString colName = sqlite3_column_name(stmt, c);
			QString colVal = QString::fromUtf8((const char*)sqlite3_column_text(stmt, c));
			row[colName] = colVal;
		}
		results += row;
		res = sqlite3_step(stmt);
	}

	sqlite3_finalize(stmt);
	return results;
}

QString DescHistory::oneColumn(QString sql, QStringList params)
{
	sqlite3_stmt* stmt;
	const char *pzTail;
	int rescode;
	QString result;

	sqlite3_prepare(db, sql.utf8(), sql.length(), &stmt, &pzTail);

	for ( uint i = 0; i < params.count(); i++ )
	{
		rescode = sqlite3_bind_text(stmt, i+1, params[i].utf8(), params[i].utf8().length(), SQLITE_TRANSIENT);
		if (rescode != SQLITE_OK)
		{
			printf("[DescHistory module] Error binding parameter %d for sql: %s\nError message: %s\n", i, sql.latin1(), sqlite3_errmsg(db));
		}
	}

	int res = sqlite3_step(stmt);
	int columns = sqlite3_column_count(stmt);
	if ( columns != 1 )
		return QString::null;

	if ( res != SQLITE_ROW )
		return QString::null;

	result = QString::fromUtf8((const char*)sqlite3_column_text(stmt, 0));

	sqlite3_finalize(stmt);
	return result;
}

void DescHistory::addEntry(UserListElement* ule)
{
	UinType uin = ule->ID("Gadu").toUInt();
	QString uid = QString::number((uint)uin);
	if ( !ule->status("Gadu").hasDescription() )
		return;

	QString desc = ule->status("Gadu").description();

	QStringList params;
	params += uid;
	QString result = oneColumn("SELECT description FROM history WHERE uid = ? ORDER BY created_on DESC LIMIT 1;", params);
	if ( result == desc )
		return;

	params.clear();
	params += uid;
	params += desc;
	call("INSERT INTO history (created_on, uid, description) VALUES (strftime('%s', 'now'), ?, ?)", params);
}

void DescHistory::statusChanged(UserListElement ule, QString protocolName, const UserStatus &oldStatus, bool massively, bool last)
{
	addEntry(&ule);
}

void DescHistory::createDefaultConfiguration()
{
	config_file.addVariable("DescHistory", "MaxDescriptions", 300);
	config_file.addVariable("DescHistory", "ExcludeMarked", true);
}

void DescHistory::mainConfigurationWindowCreated(MainConfigurationWindow *mainConfigurationWindow)
{}

DescriptionsWindow* DescHistory::createDescWindow()
{
	DescriptionsWindow* win = new DescriptionsWindow(this);
	UserListElements list = userlist->toUserListElements();
	sort(list);
	for ( uint i = 0; i < list.count(); i++ )
	{
		UserListElement ule = list[i];
		win->addContact(ule);
	}
	return win;
}

void DescHistory::showDescHistoryWin()
{
	DescriptionsWindow* win = createDescWindow();
	win->show();
}

void DescHistory::showDescHistoryWinSelected()
{
	DescriptionsWindow* win = createDescWindow();
	UserBox *activeUserBox = UserBox::activeUserBox();
	if (activeUserBox == NULL)
		return;

	if (activeUserBox->currentUserExists())
	{
		UserListElement ule = activeUserBox->currentUser();
		win->select( (UinType)ule.ID("Gadu").toUInt() );
	}
	win->show();
}

void DescHistory::sort(UserListElements &uleList)
{
	int size = uleList.size();
	for (int j = 0; j < size - 1; j++)
	{
		for (int i = 0; i < size - 1; i++)
		{
			if ( uleList[i].altNick().localeAwareCompare( uleList[i+1].altNick() ) > 0 )
			{
				//swap(uleList[i], uleList[i+1])
				UserListElement ule = uleList[i+1];
				uleList[i+1] = uleList[i];
				uleList[i] = ule;
			}
		}
	}
}
