/***************************************************************************
 *																		 *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or	 *
 *   (at your option) any later version.								   *
 *																		 *
 ***************************************************************************/

#include <qdatetimeedit.h>
#include <qpushbutton.h>
#include <qlineedit.h>
#include <qlayout.h>
#include <qlabel.h>
#include <qregexp.h>
#include <qpainter.h>
#include <qfontmetrics.h>
#include <qpopupmenu.h>
#include <qclipboard.h>
#include <qapplication.h>
#include <qcursor.h>

#include "userlist.h"
#include "misc.h"
#include "config_file.h"
#include "icons_manager.h"
#include "protocol.h"
#include "protocols_manager.h"
#include "chat_manager.h"

#include "descriptions_window.h"
#include "desc_history.h"

static const char* DescHistory_markIcon = "ApplyWindowButton";

DescHistorySource::DescHistorySource(QWidget* parent) : QComboBox(parent)
{}

void DescHistorySource::insertItem( const QString &t, UinType uin )
{
	QComboBox::insertItem(t);
	uinsMap[count()-1] = uin;
}

UinType DescHistorySource::getUin(int index)
{
	return uinsMap[index];
}

//---------------------------------------------------------------------------------

DescHistoryListItem::DescHistoryListItem(QListView* parent) : QListViewItem(parent)
{}

void DescHistoryListItem::setId(QString id)
{
	m_id = id;
}

QString DescHistoryListItem::getId()
{
	return m_id;
}

void DescHistoryListItem::setAuthorUid(QString authorUid)
{
	m_authorUid = authorUid;
}

QString DescHistoryListItem::getAuthorUid()
{
	return m_authorUid;
}

void DescHistoryListItem::setMarked(bool m)
{
	marked = m;
	if (m)
		setPixmap(1, icons_manager->loadIcon(DescHistory_markIcon));
	else
		setPixmap(1, 0);
}

bool DescHistoryListItem::getMarked()
{
	return marked;
}

void DescHistoryListItem::setup()
{
	widthChanged();

	int hg;
	QPainter p;
	QFontMetrics fm(p.font());
	hg = fm.height();
	QStringList lines = QStringList::split( "\n", text(4), true );
	setHeight(hg*lines.count());
}

void DescHistoryListItem::paintCell( QPainter * p, const QColorGroup & cg, int column, int width, int align )
{
	QFont f = p->font();
	if ( marked )
		f.setBold(true);
	else
		f.setBold(false);

	p->setFont(f);

	if ( column == 3 )
		width += 15;

	QColorGroup colors = cg;
	if ( column == 4 )
	{
		QRegExp re("(http:\\/\\/\\S+|www\\.\\S+\\.\\S+)");
		if ( re.search(text(4)) > -1 )
			colors.setColor(QColorGroup::Text, "blue");
	}

	QListViewItem::paintCell(p, colors, column, width, align);
}

//---------------------------------------------------------------------------------

DescriptionsWindow::DescriptionsWindow(DescHistory* descHistory) : QWidget(0, 0, Qt::WDestructiveClose)
{
	descHistoryObj = descHistory;

	setCaption(tr("Descriptions history list"));
	setMinimumSize(300, 200);

	int hg = config_file.readNumEntry( "DescHist", "WindowHeight", 400 );
	int wd = config_file.readNumEntry( "DescHist", "WindowWidth", 600 );
	resize(wd, hg);

	source = new DescHistorySource(this);
	source->setEditable(false);
	source->QComboBox::insertItem(tr("Marked descriptions"));
	source->QComboBox::insertItem(tr("All descriptions"));
	source->setFocusPolicy(QWidget::StrongFocus);
	connect(source, SIGNAL(activated(int)), this, SLOT(activated(int)));

	QLabel* fromLab = new QLabel(tr("From:"), this);
	fromLab->setAlignment(Qt::AlignRight|Qt::AlignVCenter);

	fromDate = new QDateTimeEdit(this);

	QLabel* toLab = new QLabel(tr("To:"), this);
	toLab->setAlignment(Qt::AlignRight|Qt::AlignVCenter);

	toDate = new QDateTimeEdit(this);

	QLabel* filterLab = new QLabel(tr("Filter:"), this);
	filterLab->setAlignment(Qt::AlignRight|Qt::AlignVCenter);

	filter = new QLineEdit(this);
	connect(filter, SIGNAL(returnPressed()), this, SLOT(refreshDescriptions()));

	applyFilter = new QPushButton(tr("Filter"), this);
	connect(applyFilter, SIGNAL(clicked()), this, SLOT(refreshDescriptions()));

	resetFilter = new QPushButton(tr("Reset"), this);
	connect(resetFilter, SIGNAL(clicked()), this, SLOT(applyReset()));

	close = new QPushButton(tr("Close"), this);
	close->setDefault(FALSE);
	connect(close, SIGNAL(clicked()), this, SLOT(close()));

	list = new QListView(this);
	list->addColumn("#");
	list->addColumn(tr("Mark"));
	list->addColumn(tr("Author"));
	list->addColumn(tr("Date"));
	list->addColumn(tr("Description"));
	list->setColumnAlignment(1, Qt::AlignCenter);
	list->setColumnAlignment(4, Qt::AlignTop);
	list->setResizeMode(QListView::LastColumn);
	list->setAllColumnsShowFocus(true);
	list->setSortColumn(0);
	list->setColumnWidth(0, 20);
	list->setColumnWidth(1, 50);
	//list->setShowSortIndicator(true);
	connect(list, SIGNAL(doubleClicked(QListViewItem *, const QPoint &, int)), this, SLOT(doubleClicked(QListViewItem *, const QPoint &, int)));

	resetValues();

	menu = new QPopupMenu(this);
	menu->insertItem(tr("Open URL from description"), this, SLOT(openDescUrl()));
	menu->insertItem(tr("Copy description to clipboard"), this, SLOT(copyDesc()));
	menu->insertSeparator();
	menu->insertItem(tr("Open Chat window with description author"), this, SLOT(openChat()));
	connect(list, SIGNAL(rightButtonClicked(QListViewItem*, const QPoint&, int)), this, SLOT(raiseMenu(QListViewItem*, const QPoint&, int)));

	// Layout
	QGridLayout *g = new QGridLayout(this, 1, 1);
	g->addMultiCellWidget(fromLab, 0, 0, 0, 1);
	g->addMultiCellWidget(fromDate, 0, 0, 2, 4);
	g->addMultiCellWidget(toLab, 0, 0, 5, 6);
	g->addMultiCellWidget(toDate, 0, 0, 7, 9);
	g->addMultiCellWidget(filterLab, 1, 1, 0, 1);
	g->addMultiCellWidget(filter, 1, 1, 2, 9);
	g->addMultiCellWidget(applyFilter, 0, 0, 11, 12);
	g->addMultiCellWidget(resetFilter, 1, 1, 11, 12);
	g->addMultiCellWidget(list, 2, 10, 0, 12);
	g->addMultiCellWidget(source, 11, 11, 0, 5);
	g->addMultiCellWidget(close, 11, 11, 10, 12);
}

DescriptionsWindow::~DescriptionsWindow()
{
	config_file.writeEntry( "DescHist", "WindowHeight", height() );
	config_file.writeEntry( "DescHist", "WindowWidth", width() );
}

void DescriptionsWindow::openDescUrl()
{
	QString desc = list->currentItem()->text(4);
	QRegExp re("(http:\\/\\/\\S+|www\\.\\S+\\.\\S+)");
	if ( re.search(desc) == -1 )
		return;

	QString link = re.cap(1);
	if ( !link.startsWith("http://") )
		link.prepend("http://");

	openWebBrowser(link);
}

void DescriptionsWindow::copyDesc()
{
	QString desc = list->currentItem()->text(4);
	QClipboard *cb = QApplication::clipboard();
	cb->setText(desc);
}

void DescriptionsWindow::openChat()
{
	QString authorId = ((DescHistoryListItem*)list->currentItem())->getAuthorUid();
	UserListElement ule = userlist->byID("Gadu", authorId);
	UserListElements uleList;
	uleList += ule;
	chat_manager->openPendingMsgs(uleList);
}

void DescriptionsWindow::raiseMenu(QListViewItem* it, const QPoint& p, int col)
{
	menu->exec(QCursor::pos());
}

void DescriptionsWindow::addContact(UserListElement ule)
{
	UinType uin = ule.ID("Gadu").toUInt();
	QString altNick = ule.altNick();
	source->insertItem(altNick, uin);
}

void DescriptionsWindow::refreshDescriptions()
{
	list->clear();

	uint from = fromDate->dateTime().toTime_t();
	uint to = toDate->dateTime().toTime_t();
	QString f = filter->text();
	int authorIdx = source->currentItem();
	if (authorIdx < 0)
	{
		printf("[DescHistory module] Can't determinate selected author to filter. This should never happen!\n");
		return;
	}

	QStringList params;
	QString sql = "SELECT * FROM history WHERE created_on BETWEEN ? AND ?";
	params += QString::number(from);
	params += QString::number(to);
	if (f.length() > 0)
	{
		sql += " AND description LIKE ?";
		f.prepend("%");
		f.append("%");
		params += f.replace(QRegExp("\\s+"), "%");
	}

	switch ( authorIdx )
	{
		case 0:
		{
			// Marked
			sql += " AND marked = 1";
			break;
		}
		case 1:
		{
			// All
			break;
		}
		default:
		{
			// Specified
			sql += " AND uid = ?";
			params += QString::number((uint)source->getUin(authorIdx));
		}
	}
	sql += " ORDER BY id DESC LIMIT 1000;";

	QValueList< QMap<QString,QString> > results = descHistoryObj->query(sql, params);
	DescHistoryListItem* item;
	int i = 1;
	int lgt = QString::number(results.count()).length();
	for ( QValueList< QMap<QString,QString> >::Iterator it = results.begin(); it != results.end(); it++ )
	{
		QMap<QString,QString> row = *it;

		QString author = getAuthorByUid((UinType)row["uid"].toUInt());
		QDateTime date;
		date.setTime_t(row["created_on"].toUInt());
		QString dateStr = date.toString(" yyyy.MM.dd  hh:mm:ss ");

		item = new DescHistoryListItem(list);
		item->setText(0, QString::number(i).rightJustify(lgt, '0'));
		item->setText(2, author);
		item->setText(3, dateStr);
		item->setText(4, row["description"]);
		item->setId(row["id"]);
		item->setAuthorUid(row["uid"]);
		item->setMarked( row["marked"].toInt() ? true : false );

		list->insertItem(item);
		item->setup();
		i++;
	}
}

QString DescriptionsWindow::getAuthorByUid(UinType uin)
{
	UserListElement ule = userlist->byID("Gadu", QString::number((uint)uin));
	return ule.altNick();
}

void DescriptionsWindow::activated(int index)
{
	refreshDescriptions();
}

void DescriptionsWindow::doubleClicked(QListViewItem* item, const QPoint &p, int col)
{
	DescHistoryListItem* it = (DescHistoryListItem*)item;
	bool mark = !it->getMarked();
	it->setMarked(mark);
	QStringList params;
	params += (mark ? "1" : "0");
	params += it->getId();
	descHistoryObj->call("UPDATE history SET marked = ? WHERE id = ?;", params);
}

void DescriptionsWindow::resetValues()
{
	QDateTime now = QDateTime::currentDateTime();
	fromDate->setDateTime( now.addYears(-1) );
	toDate->setDateTime( now );
	filter->setText("");
	source->setCurrentItem(0);
}

void DescriptionsWindow::applyReset()
{
	resetValues();
	list->clear();
}

void DescriptionsWindow::select(int index)
{
	source->setCurrentItem(0);
	refreshDescriptions();
}

void DescriptionsWindow::select(UinType uin)
{
	for ( int i = 2; i < source->count(); i++ )
	{
		if ( source->getUin(i) == uin )
		{
			source->setCurrentItem(i);
			refreshDescriptions();
			break;
		}
	}
}
